﻿using System;
using Pixiview.UI.Theme;
using Pixiview.Utils;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace Pixiview.UI
{
    public class AdaptedPage : ContentPage
    {
        public static readonly BindableProperty PageTopMarginProperty = BindableProperty.Create(
            nameof(PageTopMargin), typeof(Thickness), typeof(AdaptedPage));
        public static readonly BindableProperty PanelTopMarginProperty = BindableProperty.Create(
            nameof(PanelTopMargin), typeof(Thickness), typeof(AdaptedPage));

        public Thickness PageTopMargin
        {
            get => (Thickness)GetValue(PageTopMarginProperty);
            protected set => SetValue(PageTopMarginProperty, value);
        }
        public Thickness PanelTopMargin
        {
            get => (Thickness)GetValue(PanelTopMarginProperty);
            protected set => SetValue(PanelTopMarginProperty, value);
        }
        public Orientation CurrentOrientation { get; private set; }

        public event EventHandler Load;
        public event EventHandler Unload;
        public event EventHandler<ThicknessEventArgs> PageTopMarginChanged;
        public event EventHandler<OrientationEventArgs> OrientationChanged;

        protected static readonly bool isPhone = DeviceInfo.Idiom == DeviceIdiom.Phone;

        public AdaptedPage()
        {
            SetDynamicResource(Screen.StatusBarStyleProperty, ThemeBase.StatusBarStyle);
            Shell.SetNavBarHasShadow(this, true);
        }

        public virtual void OnLoad()
        {
            Load?.Invoke(this, EventArgs.Empty);
        }

        public virtual void OnUnload()
        {
            Unload?.Invoke(this, EventArgs.Empty);
        }

        public virtual void OnOrientationChanged(Orientation orientation)
        {
#if __IOS__
            var oldMargin = PageTopMargin;
            Thickness newMargin;
            switch (orientation)
            {
                case Orientation.Portrait:
                    newMargin = AppShell.TotalBarOffset;
                    break;
                case Orientation.PortraitUpsideDown:
                case Orientation.Unknown:
                case Orientation.LandscapeLeft:
                case Orientation.LandscapeRight:
                default:
                    if (StyleDefinition.IsFullscreenDevice)
                    {
                        newMargin = AppShell.NavigationBarOffset;
                    }
                    else
                    {
                        newMargin = isPhone ? StyleDefinition.TopOffset32 : AppShell.TotalBarOffset;
                    }
                    break;
            }
            if (oldMargin != newMargin)
            {
                PageTopMargin = newMargin;
                OnPageTopMarginChanged(oldMargin, newMargin);
            }
#endif
            OrientationChanged?.Invoke(this, new OrientationEventArgs { CurrentOrientation = orientation });
        }

#if __IOS__
        protected override void OnSizeAllocated(double width, double height)
        {
            base.OnSizeAllocated(width, height);

            if (PageTopMargin == default)
            {
                if (StyleDefinition.IsFullscreenDevice)
                {
                    PageTopMargin = width > height ?
                        AppShell.NavigationBarOffset :
                        AppShell.TotalBarOffset;
                }
                else if (isPhone)
                {
                    PageTopMargin = width > height ?
                        StyleDefinition.TopOffset32 :
                        AppShell.TotalBarOffset;
                }
                else
                {
                    PageTopMargin = AppShell.TotalBarOffset;
                }
            }
        }
#endif

        public virtual void OnPageTopMarginChanged(Thickness old, Thickness @new)
        {
            PageTopMarginChanged?.Invoke(this, new ThicknessEventArgs
            {
                OldMargin = old,
                NewMargin = @new
            });
        }

        public virtual void OnPanelTopMarginChanged(Thickness old, Thickness @new)
        {
        }

        public void InitOrientation(Orientation orientation)
        {
            CurrentOrientation = orientation;
        }

        protected void Start(Action action)
        {
            if (Tap.IsBusy)
            {
                App.DebugPrint("gesture recognizer is now busy...");
                return;
            }
            using (Tap.Start())
            {
                action();
            }
        }

        private class Tap : IDisposable
        {
            public static bool IsBusy => _instance?.isBusy == true;

            private static readonly object sync = new object();
            private static readonly Tap _instance = new Tap();

            private Tap() { }

            public static Tap Start()
            {
                lock (sync)
                {
                    _instance.isBusy = true;
                }
                return _instance;
            }

            private volatile bool isBusy = false;

            public void Dispose()
            {
                lock (sync)
                {
                    isBusy = false;
                }
            }
        }
    }

    public class ThicknessEventArgs : EventArgs
    {
        public Thickness OldMargin { get; set; }
        public Thickness NewMargin { get; set; }
    }

    public class OrientationEventArgs : EventArgs
    {
        public Orientation CurrentOrientation { get; set; }
    }

    public enum Orientation
    {
        Unknown,
        Portrait,
        PortraitUpsideDown,
        LandscapeLeft,
        LandscapeRight
    }
}
